@@ -1,19 +1,22 @@
 # -*- encoding: utf-8 -*-
 import functools
 import logging
+import json
 import math
 import multiprocessing
 from queue import Empty
+import time
 import traceback
-from typing import Optional
-
-import numpy as np
-import pynisher
-from smac.tae.execute_ta_run import StatusType, BudgetExhaustedException, \
-    TAEAbortException
-from smac.tae.execute_func import AbstractTAFunc
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
 
 from ConfigSpace import Configuration
+import numpy as np
+import pynisher
+from smac.runhistory.runhistory import RunInfo, RunValue
+from smac.stats.stats import Stats
+from smac.tae import StatusType, TAEAbortException
+from smac.tae.execute_func import AbstractTAFunc
+
 from sklearn.model_selection._split import _RepeatedSplits, BaseShuffleSplit,\
     BaseCrossValidator
 from autosklearn.metrics import Scorer
@@ -21,9 +24,17 @@ from autosklearn.metrics import Scorer
 import autosklearn.evaluation.train_evaluator
 import autosklearn.evaluation.test_evaluator
 import autosklearn.evaluation.util
+from autosklearn.evaluation.train_evaluator import TYPE_ADDITIONAL_INFO
+from autosklearn.util.backend import Backend
+from autosklearn.util.logging_ import PickableLoggerAdapter, get_named_client_logger
+from autosklearn.util.parallel import preload_modules
 
 
-def fit_predict_try_except_decorator(ta, queue, cost_for_crash, **kwargs):
+def fit_predict_try_except_decorator(
+        ta: Callable,
+        queue: multiprocessing.Queue,
+        cost_for_crash: float,
+        **kwargs: Any) -> None:
 
     try:
         return ta(queue=queue, **kwargs)
@@ -35,14 +46,34 @@ def fit_predict_try_except_decorator(ta, queue, cost_for_crash, **kwargs):
         exception_traceback = traceback.format_exc()
         error_message = repr(e)
 
+        # Printing stuff to stdout just in case the queue doesn't work, which happened with the
+        # following traceback:
+        #     File "auto-sklearn/autosklearn/evaluation/__init__.py", line 29, in fit_predict_try_except_decorator  # noqa E501
+        #     return ta(queue=queue, **kwargs)
+        #     File "auto-sklearn/autosklearn/evaluation/train_evaluator.py", line 1067, in eval_holdout  # noqa E501
+        #     evaluator.fit_predict_and_loss(iterative=iterative)
+        #     File "auto-sklearn/autosklearn/evaluation/train_evaluator.py", line 616, in fit_predict_and_loss,  # noqa E501
+        #     status=status
+        #     File "auto-sklearn/autosklearn/evaluation/abstract_evaluator.py", line 320, in finish_up  # noqa E501
+        #     self.queue.put(rval_dict)
+        #     File "miniconda/3-4.5.4/envs/autosklearn/lib/python3.7/multiprocessing/queues.py", line 87, in put  # noqa E501
+        #     self._start_thread()
+        #     File "miniconda/3-4.5.4/envs/autosklearn/lib/python3.7/multiprocessing/queues.py", line 170, in _start_thread  # noqa E501
+        #     self._thread.start()
+        #     File "miniconda/3-4.5.4/envs/autosklearn/lib/python3.7/threading.py", line 847, in start  # noqa E501
+        #     RuntimeError: can't start new thread
+        print("Exception handling in `fit_predict_try_except_decorator`: "
+              "traceback: %s \nerror message: %s" % (exception_traceback, error_message))
+
         queue.put({'loss': cost_for_crash,
                    'additional_run_info': {'traceback': exception_traceback,
                                            'error': error_message},
                    'status': StatusType.CRASHED,
-                   'final_queue_element': True})
+                   'final_queue_element': True}, block=True)
+        queue.close()
 
 
-def get_cost_of_crash(metric):
+def get_cost_of_crash(metric: Scorer) -> float:
 
     # The metric must always be defined to extract optimum/worst
     if not isinstance(metric, Scorer):
@@ -61,16 +92,46 @@ def get_cost_of_crash(metric):
     return worst_possible_result
 
 
+def _encode_exit_status(exit_status: Union[str, int, Type[BaseException]]
+                        ) -> Union[str, int]:
+    try:
+        # If it can be dumped, then it is int
+        exit_status = cast(int, exit_status)
+        json.dumps(exit_status)
+        return exit_status
+    except (TypeError, OverflowError):
+        return str(exit_status)
+
+
 # TODO potentially log all inputs to this class to pickle them in order to do
 # easier debugging of potential crashes
 class ExecuteTaFuncWithQueue(AbstractTAFunc):
 
-    def __init__(self, backend, autosklearn_seed, resampling_strategy, metric,
-                 logger, initial_num_run=1, stats=None, runhistory=None,
-                 run_obj='quality', par_factor=1, all_scoring_functions=False,
-                 output_y_hat_optimization=True, include=None, exclude=None,
-                 memory_limit=None, disable_file_output=False, init_params=None,
-                 budget_type=None, **resampling_strategy_args):
+    def __init__(
+        self,
+        backend: Backend,
+        autosklearn_seed: int,
+        resampling_strategy: Union[str, BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit],
+        metric: Scorer,
+        cost_for_crash: float,
+        abort_on_first_run_crash: bool,
+        port: int,
+        pynisher_context: str,
+        initial_num_run: int = 1,
+        stats: Optional[Stats] = None,
+        run_obj: str = 'quality',
+        par_factor: int = 1,
+        scoring_functions: Optional[List[Scorer]] = None,
+        output_y_hat_optimization: bool = True,
+        include: Optional[List[str]] = None,
+        exclude: Optional[List[str]] = None,
+        memory_limit: Optional[int] = None,
+        disable_file_output: bool = False,
+        init_params: Optional[Dict[str, Any]] = None,
+        budget_type: Optional[str] = None,
+        ta: Optional[Callable] = None,
+        **resampling_strategy_args: Any,
+    ):
 
         if resampling_strategy == 'holdout':
             eval_function = autosklearn.evaluation.train_evaluator.eval_holdout
@@ -78,13 +139,9 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
             eval_function = autosklearn.evaluation.train_evaluator.eval_iterative_holdout
         elif resampling_strategy == 'cv-iterative-fit':
             eval_function = autosklearn.evaluation.train_evaluator.eval_iterative_cv
-        elif resampling_strategy == 'cv' or (
-             isinstance(resampling_strategy, type) and (
-                issubclass(resampling_strategy, BaseCrossValidator) or
-                issubclass(resampling_strategy, _RepeatedSplits) or
-                issubclass(resampling_strategy, BaseShuffleSplit)
-                )
-             ):
+        elif resampling_strategy == 'cv' or isinstance(resampling_strategy, (
+            BaseCrossValidator, _RepeatedSplits, BaseShuffleSplit)
+        ):
             eval_function = autosklearn.evaluation.train_evaluator.eval_cv
         elif resampling_strategy == 'partial-cv':
             eval_function = autosklearn.evaluation.train_evaluator.eval_partial_cv
@@ -97,7 +154,7 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
             raise ValueError('Unknown resampling strategy %s' %
                              resampling_strategy)
 
-        self.worst_possible_result = get_cost_of_crash(metric)
+        self.worst_possible_result = cost_for_crash
 
         eval_function = functools.partial(
             fit_predict_try_except_decorator,
@@ -108,20 +165,20 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
         super().__init__(
             ta=eval_function,
             stats=stats,
-            runhistory=runhistory,
             run_obj=run_obj,
             par_factor=par_factor,
             cost_for_crash=self.worst_possible_result,
+            abort_on_first_run_crash=abort_on_first_run_crash,
         )
 
         self.backend = backend
         self.autosklearn_seed = autosklearn_seed
         self.resampling_strategy = resampling_strategy
-        self.num_run = initial_num_run
+        self.initial_num_run = initial_num_run
         self.metric = metric
         self.resampling_strategy = resampling_strategy
         self.resampling_strategy_args = resampling_strategy_args
-        self.all_scoring_functions = all_scoring_functions
+        self.scoring_functions = scoring_functions
         # TODO deactivate output_y_hat_optimization and let the respective evaluator decide
         self.output_y_hat_optimization = output_y_hat_optimization
         self.include = include
@@ -129,7 +186,6 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
         self.disable_file_output = disable_file_output
         self.init_params = init_params
         self.budget_type = budget_type
-        self.logger = logger
 
         if memory_limit is not None:
             memory_limit = int(math.ceil(memory_limit))
@@ -145,55 +201,47 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
         else:
             self._get_test_loss = False
 
-    def start(self, config: Configuration,
-              instance: Optional[str],
-              cutoff: float = None,
-              seed: int = 12345,
-              budget: float = 0.0,
-              instance_specific: Optional[str] = None,
-              capped: bool = False):
+        self.port = port
+        self.pynisher_context = pynisher_context
+        if self.port is None:
+            self.logger: Union[logging.Logger, PickableLoggerAdapter] = logging.getLogger("TAE")
+        else:
+            self.logger = get_named_client_logger(
+                name="TAE",
+                port=self.port,
+            )
+
+    def run_wrapper(
+        self,
+        run_info: RunInfo,
+    ) -> Tuple[RunInfo, RunValue]:
         """
-        wrapper function for ExecuteTARun.start() to cap the target algorithm
+        wrapper function for ExecuteTARun.run_wrapper() to cap the target algorithm
         runtime if it would run over the total allowed runtime.
 
         Parameters
         ----------
-            config : Configuration
-                mainly a dictionary param -> value
-            instance : string
-                problem instance
-            cutoff : float
-                runtime cutoff
-            seed : int
-                random seed
-            budget : float
-                A positive, real-valued number representing an arbitrary limit to the target
-                algorithm. Handled by the target algorithm internally
-            instance_specific: str
-                instance specific information (e.g., domain file or solution)
-            capped: bool
-                if true and status is StatusType.TIMEOUT,
-                uses StatusType.CAPPED
+        run_info : RunInfo
+            Object that contains enough information to execute a configuration run in
+            isolation.
         Returns
         -------
-            status: enum of StatusType (int)
-                {SUCCESS, TIMEOUT, CRASHED, ABORT}
-            cost: float
-                cost/regret/quality (float) (None, if not returned by TA)
-            runtime: float
-                runtime (None if not returned by TA)
-            additional_info: dict
-                all further additional run information
+        RunInfo:
+            an object containing the configuration launched
+        RunValue:
+            Contains information about the status/performance of config
         """
         if self.budget_type is None:
-            if budget != 0:
-                raise ValueError('If budget_type is None, budget must be.0, but is %f' % budget)
+            if run_info.budget != 0:
+                raise ValueError(
+                    'If budget_type is None, budget must be.0, but is %f' % run_info.budget
+                )
         else:
-            if budget == 0:
-                budget = 100
-            elif budget <= 0 or budget > 100:
+            if run_info.budget == 0:
+                run_info = run_info._replace(budget=100)
+            elif run_info.budget <= 0 or run_info.budget > 100:
                 raise ValueError('Illegal value for budget, must be >0 and <=100, but is %f' %
-                                 budget)
+                                 run_info.budget)
             if self.budget_type not in ('subsample', 'iterations', 'mixed'):
                 raise ValueError("Illegal value for budget type, must be one of "
                                  "('subsample', 'iterations', 'mixed'), but is : %s" %
@@ -201,24 +249,49 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
 
         remaining_time = self.stats.get_remaing_time_budget()
 
-        if remaining_time - 5 < cutoff:
-            cutoff = int(remaining_time - 5)
+        if remaining_time - 5 < run_info.cutoff:
+            run_info = run_info._replace(cutoff=int(remaining_time - 5))
 
-        if cutoff < 1.0:
-            raise BudgetExhaustedException()
-        cutoff = int(np.ceil(cutoff))
+        config_id = (
+            run_info.config if isinstance(run_info.config, int) else run_info.config.config_id
+        )
 
-        return super().start(config=config, instance=instance, cutoff=cutoff,
-                             seed=seed, instance_specific=instance_specific,
-                             capped=capped, budget=budget)
+        if run_info.cutoff < 1.0:
+            self.logger.info("Not starting configuration %d because time is up" % config_id)
+            return run_info, RunValue(
+                status=StatusType.STOP,
+                cost=self.worst_possible_result,
+                time=0.0,
+                additional_info={},
+                starttime=time.time(),
+                endtime=time.time(),
+            )
+        elif (
+            run_info.cutoff != int(np.ceil(run_info.cutoff))
+            and not isinstance(run_info.cutoff, int)
+        ):
+            run_info = run_info._replace(cutoff=int(np.ceil(run_info.cutoff)))
 
-    def run(self, config, instance=None,
-            cutoff=None,
-            seed=12345,
-            budget=0.0,
-            instance_specific=None):
+        self.logger.info("Starting to evaluate configuration %d" % config_id)
+        return super().run_wrapper(run_info=run_info)
 
-        queue = multiprocessing.Queue()
+    def run(
+        self,
+        config: Configuration,
+        instance: Optional[str] = None,
+        cutoff: Optional[float] = None,
+        seed: int = 12345,
+        budget: float = 0.0,
+        instance_specific: Optional[str] = None,
+    ) -> Tuple[StatusType, float, float, Dict[str, Union[int, float, str, Dict, List, Tuple]]]:
+
+        # Additional information of each of the tae executions
+        # Defined upfront for mypy
+        additional_run_info: TYPE_ADDITIONAL_INFO = {}
+
+        context = multiprocessing.get_context(self.pynisher_context)
+        preload_modules(context)
+        queue = context.Queue()
 
         if not (instance_specific is None or instance_specific == '0'):
             raise ValueError(instance_specific)
@@ -226,19 +299,35 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
         if self.init_params is not None:
             init_params.update(self.init_params)
 
+        if self.port is None:
+            logger: Union[logging.Logger, PickableLoggerAdapter] = logging.getLogger("pynisher")
+        else:
+            logger = get_named_client_logger(
+                name="pynisher",
+                port=self.port,
+            )
         arguments = dict(
-            logger=logging.getLogger("pynisher"),
+            logger=logger,
             wall_time_in_s=cutoff,
             mem_in_mb=self.memory_limit,
+            capture_output=True,
+            context=context,
         )
+
+        if isinstance(config, int):
+            num_run = self.initial_num_run
+        else:
+            num_run = config.config_id + self.initial_num_run
+
         obj_kwargs = dict(
             queue=queue,
             config=config,
             backend=self.backend,
+            port=self.port,
             metric=self.metric,
             seed=self.autosklearn_seed,
-            num_run=self.num_run,
-            all_scoring_functions=self.all_scoring_functions,
+            num_run=num_run,
+            scoring_functions=self.scoring_functions,
             output_y_hat_optimization=self.output_y_hat_optimization,
             include=self.include,
             exclude=self.exclude,
@@ -253,11 +342,19 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
             obj_kwargs['resampling_strategy'] = self.resampling_strategy
             obj_kwargs['resampling_strategy_args'] = self.resampling_strategy_args
 
-        obj = pynisher.enforce_limits(**arguments)(self.ta)
-        obj(**obj_kwargs)
+        try:
+            obj = pynisher.enforce_limits(**arguments)(self.ta)
+            obj(**obj_kwargs)
+        except Exception as e:
+            exception_traceback = traceback.format_exc()
+            error_message = repr(e)
+            additional_run_info.update({
+                'traceback': exception_traceback,
+                'error': error_message
+            })
+            return StatusType.CRASHED, self.worst_possible_result, 0.0, additional_run_info
 
-        if obj.exit_status in (pynisher.TimeoutException,
-                               pynisher.MemorylimitException):
+        if obj.exit_status in (pynisher.TimeoutException, pynisher.MemorylimitException):
             # Even if the pynisher thinks that a timeout or memout occured,
             # it can be that the target algorithm wrote something into the queue
             #  - then we treat it as a succesful run
@@ -267,6 +364,11 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
                 status = info[-1]['status']
                 additional_run_info = info[-1]['additional_run_info']
 
+                if obj.stdout:
+                    additional_run_info['subprocess_stdout'] = obj.stdout
+                if obj.stderr:
+                    additional_run_info['subprocess_stderr'] = obj.stderr
+
                 if obj.exit_status is pynisher.TimeoutException:
                     additional_run_info['info'] = 'Run stopped because of timeout.'
                 elif obj.exit_status is pynisher.MemorylimitException:
@@ -285,8 +387,7 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
                 elif obj.exit_status is pynisher.MemorylimitException:
                     status = StatusType.MEMOUT
                     additional_run_info = {
-                        'error': 'Memout (used more than %d MB).' %
-                                 self.memory_limit
+                        "error": "Memout (used more than {} MB).".format(self.memory_limit)
                     }
                 else:
                     raise ValueError(obj.exit_status)
@@ -297,7 +398,11 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
             status = StatusType.ABORT
             cost = self.worst_possible_result
             additional_run_info = {'error': 'Your configuration of '
-                                            'auto-sklearn does not work!'}
+                                            'auto-sklearn does not work!',
+                                   'exit_status': _encode_exit_status(obj.exit_status),
+                                   'subprocess_stdout': obj.stdout,
+                                   'subprocess_stderr': obj.stderr,
+                                   }
 
         else:
             try:
@@ -315,9 +420,18 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
                                                   'because the pynisher exit ' \
                                                   'status %s is unknown.' % \
                                                   str(obj.exit_status)
+                    additional_run_info['exit_status'] = _encode_exit_status(obj.exit_status)
+                    additional_run_info['subprocess_stdout'] = obj.stdout
+                    additional_run_info['subprocess_stderr'] = obj.stderr
             except Empty:
                 info = None
-                additional_run_info = {'error': 'Result queue is empty'}
+                additional_run_info = {
+                    'error': 'Result queue is empty',
+                    'exit_status': _encode_exit_status(obj.exit_status),
+                    'subprocess_stdout': obj.stdout,
+                    'subprocess_stderr': obj.stderr,
+                    'exitcode': obj.exitcode
+                }
                 status = StatusType.CRASHED
                 cost = self.worst_possible_result
 
@@ -371,13 +485,14 @@ class ExecuteTaFuncWithQueue(AbstractTAFunc):
 
         if isinstance(config, int):
             origin = 'DUMMY'
+            config_id = config
         else:
             origin = getattr(config, 'origin', 'UNKNOWN')
+            config_id = config.config_id
         additional_run_info['configuration_origin'] = origin
 
         runtime = float(obj.wall_clock_time)
-        self.num_run += 1
 
         autosklearn.evaluation.util.empty_queue(queue)
-
+        self.logger.info("Finished evaluating configuration %d" % config_id)
         return status, cost, runtime, additional_run_info
