@@ -1,13 +1,19 @@
-import os
+import logging
+import multiprocessing
 import time
 import warnings
-from collections import namedtuple
+from typing import Any, Dict, List, Optional, TextIO, Tuple, Type, Union, cast
 
-import lockfile
 import numpy as np
+
+from sklearn.base import BaseEstimator
 from sklearn.dummy import DummyClassifier, DummyRegressor
 from sklearn.ensemble import VotingClassifier, VotingRegressor
 
+from smac.tae import StatusType
+
+from threadpoolctl import threadpool_limits
+
 import autosklearn.pipeline.classification
 import autosklearn.pipeline.regression
 from autosklearn.constants import (
@@ -15,12 +21,14 @@ from autosklearn.constants import (
     REGRESSION_TASKS,
     MULTILABEL_CLASSIFICATION,
     MULTICLASS_CLASSIFICATION,
+    MULTIOUTPUT_REGRESSION
 )
 from autosklearn.pipeline.implementations.util import (
     convert_multioutput_multiclass_to_multilabel
 )
-from autosklearn.metrics import calculate_score, CLASSIFICATION_METRICS
-from autosklearn.util.logging_ import get_logger
+from autosklearn.metrics import calculate_loss, Scorer
+from autosklearn.util.backend import Backend
+from autosklearn.util.logging_ import PicklableClientLogger, get_named_client_logger
 
 from ConfigSpace import Configuration
 
@@ -29,79 +37,135 @@ __all__ = [
     'AbstractEvaluator'
 ]
 
-WriteTask = namedtuple('WriteTask', ['lock', 'writer', 'args'])
+
+# General TYPE definitions for numpy
+TYPE_ADDITIONAL_INFO = Dict[str, Union[int, float, str, Dict, List, Tuple]]
 
 
 class MyDummyClassifier(DummyClassifier):
-    def __init__(self, configuration, random_state, init_params=None):
-        self.configuration = configuration
-        if configuration == 1:
+    def __init__(
+        self,
+        config: Configuration,
+        random_state: np.random.RandomState,
+        init_params: Optional[Dict[str, Any]] = None,
+        dataset_properties: Dict[str, Any] = {},
+        include: Optional[List[str]] = None,
+        exclude: Optional[List[str]] = None,
+    ):
+        self.config = config
+        if config == 1:
             super(MyDummyClassifier, self).__init__(strategy="uniform")
         else:
             super(MyDummyClassifier, self).__init__(strategy="most_frequent")
+        self.random_state = random_state
+        self.init_params = init_params
+        self.dataset_properties = dataset_properties
+        self.include = include
+        self.exclude = exclude
 
-    def pre_transform(self, X, y, fit_params=None):  # pylint: disable=R0201
+    def pre_transform(
+        self,
+        X: np.ndarray,
+        y: np.ndarray,
+        fit_params: Optional[Dict[str, Any]] = None
+    ) -> Tuple[np.ndarray, Dict[str, Any]]:  # pylint: disable=R0201
         if fit_params is None:
             fit_params = {}
         return X, fit_params
 
-    def fit(self, X, y, sample_weight=None):
+    def fit(self, X: np.ndarray, y: np.ndarray,
+            sample_weight: Optional[Union[np.ndarray, List]] = None
+            ) -> DummyClassifier:
         return super(MyDummyClassifier, self).fit(np.ones((X.shape[0], 1)), y,
                                                   sample_weight=sample_weight)
 
-    def fit_estimator(self, X, y, fit_params=None):
+    def fit_estimator(self, X: np.ndarray, y: np.ndarray,
+                      fit_params: Optional[Dict[str, Any]] = None) -> DummyClassifier:
         return self.fit(X, y)
 
-    def predict_proba(self, X, batch_size=1000):
+    def predict_proba(self, X: np.ndarray, batch_size: int = 1000
+                      ) -> np.ndarray:
         new_X = np.ones((X.shape[0], 1))
         probas = super(MyDummyClassifier, self).predict_proba(new_X)
         probas = convert_multioutput_multiclass_to_multilabel(probas).astype(
             np.float32)
         return probas
 
-    def estimator_supports_iterative_fit(self):  # pylint: disable=R0201
+    def estimator_supports_iterative_fit(self) -> bool:  # pylint: disable=R0201
         return False
 
-    def get_additional_run_info(self):  # pylint: disable=R0201
+    def get_additional_run_info(self) -> Optional[TYPE_ADDITIONAL_INFO]:  # pylint: disable=R0201
         return None
 
 
 class MyDummyRegressor(DummyRegressor):
-    def __init__(self, configuration, random_state, init_params=None):
-        self.configuration = configuration
-        if configuration == 1:
+    def __init__(
+        self,
+        config: Configuration,
+        random_state: np.random.RandomState,
+        init_params: Optional[Dict[str, Any]] = None,
+        dataset_properties: Dict[str, Any] = {},
+        include: Optional[List[str]] = None,
+        exclude: Optional[List[str]] = None,
+    ):
+        self.config = config
+        if config == 1:
             super(MyDummyRegressor, self).__init__(strategy='mean')
         else:
             super(MyDummyRegressor, self).__init__(strategy='median')
+        self.random_state = random_state
+        self.init_params = init_params
+        self.dataset_properties = dataset_properties
+        self.include = include
+        self.exclude = exclude
 
-    def pre_transform(self, X, y, fit_params=None):
+    def pre_transform(
+        self,
+        X: np.ndarray,
+        y: np.ndarray,
+        fit_params: Optional[Dict[str, Any]] = None
+    ) -> Tuple[np.ndarray, Dict[str, Any]]:  # pylint: disable=R0201
         if fit_params is None:
             fit_params = {}
         return X, fit_params
 
-    def fit(self, X, y, sample_weight=None):
+    def fit(self, X: np.ndarray, y: np.ndarray,
+            sample_weight: Optional[Union[np.ndarray, List]] = None
+            ) -> DummyRegressor:
         return super(MyDummyRegressor, self).fit(np.ones((X.shape[0], 1)), y,
                                                  sample_weight=sample_weight)
 
-    def fit_estimator(self, X, y, fit_params=None):
+    def fit_estimator(self, X: np.ndarray, y: np.ndarray,
+                      fit_params: Optional[Dict[str, Any]] = None) -> DummyRegressor:
         return self.fit(X, y)
 
-    def predict(self, X, batch_size=1000):
+    def predict(self, X: np.ndarray, batch_size: int = 1000) -> np.ndarray:
         new_X = np.ones((X.shape[0], 1))
         return super(MyDummyRegressor, self).predict(new_X).astype(np.float32)
 
-    def estimator_supports_iterative_fit(self):  # pylint: disable=R0201
+    def estimator_supports_iterative_fit(self) -> bool:  # pylint: disable=R0201
         return False
 
-    def get_additional_run_info(self):  # pylint: disable=R0201
+    def get_additional_run_info(self) -> Optional[TYPE_ADDITIONAL_INFO]:  # pylint: disable=R0201
         return None
 
 
-def _fit_and_suppress_warnings(logger, model, X, y):
-    def send_warnings_to_log(message, category, filename, lineno,
-                             file=None, line=None):
-        logger.debug('%s:%s: %s:%s',
-                     filename, lineno, category.__name__, message)
+def _fit_and_suppress_warnings(
+    logger: Union[logging.Logger, PicklableClientLogger],
+    model: BaseEstimator,
+    X: np.ndarray,
+    y: np.ndarray
+) -> BaseEstimator:
+    def send_warnings_to_log(
+        message: Union[Warning, str],
+        category: Type[Warning],
+        filename: str,
+        lineno: int,
+        file: Optional[TextIO] = None,
+        line: Optional[str] = None,
+    ) -> None:
+        logger.debug('%s:%s: %s:%s' %
+                     (filename, lineno, str(category), message))
         return
 
     with warnings.catch_warnings():
@@ -112,23 +176,33 @@ def _fit_and_suppress_warnings(logger, model, X, y):
 
 
 class AbstractEvaluator(object):
-    def __init__(self, backend, queue, metric,
-                 configuration=None,
-                 all_scoring_functions=False,
-                 seed=1,
-                 output_y_hat_optimization=True,
-                 num_run=None,
-                 include=None,
-                 exclude=None,
-                 disable_file_output=False,
-                 init_params=None,
-                 budget=None,
-                 budget_type=None):
+    def __init__(
+        self,
+        backend: Backend,
+        queue: multiprocessing.Queue,
+        metric: Scorer,
+        port: Optional[int],
+        configuration: Optional[Union[int, Configuration]] = None,
+        scoring_functions: Optional[List[Scorer]] = None,
+        seed: int = 1,
+        output_y_hat_optimization: bool = True,
+        num_run: Optional[int] = None,
+        include: Optional[List[str]] = None,
+        exclude: Optional[List[str]] = None,
+        disable_file_output: Union[bool, List[str]] = False,
+        init_params: Optional[Dict[str, Any]] = None,
+        budget: Optional[float] = None,
+        budget_type: Optional[str] = None,
+    ):
+
+        # Limit the number of threads that numpy uses
+        threadpool_limits(limits=1)
 
         self.starttime = time.time()
 
         self.configuration = configuration
         self.backend = backend
+        self.port = port
         self.queue = queue
 
         self.datamanager = self.backend.load_datamanager()
@@ -145,10 +219,10 @@ class AbstractEvaluator(object):
         self.seed = seed
 
         self.output_y_hat_optimization = output_y_hat_optimization
-        self.all_scoring_functions = all_scoring_functions
+        self.scoring_functions = scoring_functions
 
         if isinstance(disable_file_output, (bool, list)):
-            self.disable_file_output = disable_file_output
+            self.disable_file_output: Union[bool, List[str]] = disable_file_output
         else:
             raise ValueError('disable_file_output should be either a bool or a list')
 
@@ -166,21 +240,11 @@ class AbstractEvaluator(object):
                 self.model_class = autosklearn.pipeline.classification.SimpleClassificationPipeline
             self.predict_function = self._predict_proba
 
-        categorical_mask = []
-        for feat in self.datamanager.feat_type:
-            if feat.lower() == 'numerical':
-                categorical_mask.append(False)
-            elif feat.lower() == 'categorical':
-                categorical_mask.append(True)
-            else:
-                raise ValueError(feat)
-        if np.sum(categorical_mask) > 0:
-            self._init_params = {
-                'data_preprocessing:categorical_features':
-                    categorical_mask
-            }
-        else:
-            self._init_params = {}
+        self._init_params = {
+            'data_preprocessing:feat_type':
+                self.datamanager.feat_type
+        }
+
         if init_params is not None:
             self._init_params.update(init_params)
 
@@ -190,26 +254,43 @@ class AbstractEvaluator(object):
 
         logger_name = '%s(%d):%s' % (self.__class__.__name__.split('.')[-1],
                                      self.seed, self.datamanager.name)
-        self.logger = get_logger(logger_name)
 
-        self.Y_optimization = None
+        if self.port is None:
+            self.logger = logging.getLogger(__name__)
+        else:
+            self.logger = get_named_client_logger(
+                name=logger_name,
+                port=self.port,
+            )
+
+        self.Y_optimization: Optional[Union[List, np.ndarray]] = None
         self.Y_actual_train = None
 
         self.budget = budget
         self.budget_type = budget_type
 
-    def _get_model(self):
+        # Please mypy to prevent not defined attr
+        self.model = self._get_model()
+
+    def _get_model(self) -> BaseEstimator:
         if not isinstance(self.configuration, Configuration):
-            model = self.model_class(configuration=self.configuration,
+            model = self.model_class(config=self.configuration,
                                      random_state=self.seed,
                                      init_params=self._init_params)
         else:
-            dataset_properties = {
-                'task': self.task_type,
-                'sparse': self.datamanager.info['is_sparse'] == 1,
-                'multilabel': self.task_type == MULTILABEL_CLASSIFICATION,
-                'multiclass': self.task_type == MULTICLASS_CLASSIFICATION,
-            }
+            if self.task_type in REGRESSION_TASKS:
+                dataset_properties = {
+                    'task': self.task_type,
+                    'sparse': self.datamanager.info['is_sparse'] == 1,
+                    'multioutput': self.task_type == MULTIOUTPUT_REGRESSION,
+                }
+            else:
+                dataset_properties = {
+                    'task': self.task_type,
+                    'sparse': self.datamanager.info['is_sparse'] == 1,
+                    'multilabel': self.task_type == MULTILABEL_CLASSIFICATION,
+                    'multiclass': self.task_type == MULTICLASS_CLASSIFICATION,
+                }
             model = self.model_class(config=self.configuration,
                                      dataset_properties=dataset_properties,
                                      random_state=self.seed,
@@ -218,46 +299,47 @@ class AbstractEvaluator(object):
                                      init_params=self._init_params)
         return model
 
-    def _loss(self, y_true, y_hat, all_scoring_functions=None):
-        """Auto-sklearn follows a minimization goal, so the make_scorer
-        sign is used as a guide to obtain the value to reduce.
+    def _loss(self, y_true: np.ndarray, y_hat: np.ndarray,
+              scoring_functions: Optional[List[Scorer]] = None
+              ) -> Union[float, Dict[str, float]]:
+        """Auto-sklearn follows a minimization goal.
+        The calculate_loss internally translate a score function to
+        a minimization problem.
 
-        On this regard, to optimize a metric:
-            1- score is calculared with calculate_score, with the caveat, that if
-            for the metric greater is not better, a negative score is returned.
-            2- the err (the optimization goal) is then:
-                optimum - (metric.sign * actual_score)
-                For accuracy for example: optimum(1) - (+1 * actual score)
-                For logloss for example: optimum(0) - (-1 * actual score)
+        For a dummy prediction, the worst result is assumed.
+
+        Parameters
+        ----------
+            y_true
         """
-        all_scoring_functions = (
-            self.all_scoring_functions
-            if all_scoring_functions is None
-            else all_scoring_functions
+        scoring_functions = (
+            self.scoring_functions
+            if scoring_functions is None
+            else scoring_functions
         )
         if not isinstance(self.configuration, Configuration):
-            if all_scoring_functions:
-                return {self.metric: 1.0}
+            if scoring_functions:
+                return {self.metric.name: self.metric._worst_possible_result}
             else:
-                return 1.0
+                return self.metric._worst_possible_result
 
-        score = calculate_score(
+        return calculate_loss(
             y_true, y_hat, self.task_type, self.metric,
-            all_scoring_functions=all_scoring_functions)
+            scoring_functions=scoring_functions)
 
-        if hasattr(score, '__len__'):
-            # TODO: instead of using self.metric, it should use all metrics given by key.
-            # But now this throws error...
-
-            err = {key: metric._optimum - score[key] for key, metric in
-                   CLASSIFICATION_METRICS.items() if key in score}
-        else:
-            err = self.metric._optimum - score
-
-        return err
-
-    def finish_up(self, loss, train_loss,  opt_pred, valid_pred, test_pred,
-                  additional_run_info, file_output, final_call, status):
+    def finish_up(
+        self,
+        loss: Union[Dict[str, float], float],
+        train_loss: Optional[Union[float, Dict[str, float]]],
+        opt_pred: np.ndarray,
+        valid_pred: np.ndarray,
+        test_pred: np.ndarray,
+        additional_run_info: Optional[TYPE_ADDITIONAL_INFO],
+        file_output: bool,
+        final_call: bool,
+        status: StatusType,
+    ) -> Tuple[float, Union[float, Dict[str, float]], int,
+               Dict[str, Union[str, int, float, Dict, List, Tuple]]]:
         """This function does everything necessary after the fitting is done:
 
         * predicting
@@ -269,19 +351,19 @@ class AbstractEvaluator(object):
         self.duration = time.time() - self.starttime
 
         if file_output:
-            loss_, additional_run_info_ = self.file_output(
+            file_out_loss, additional_run_info_ = self.file_output(
                 opt_pred, valid_pred, test_pred,
             )
         else:
-            loss_ = None
+            file_out_loss = None
             additional_run_info_ = {}
 
         validation_loss, test_loss = self.calculate_auxiliary_losses(
             valid_pred, test_pred,
         )
 
-        if loss_ is not None:
-            return self.duration, loss_, self.seed, additional_run_info_
+        if file_out_loss is not None:
+            return self.duration, file_out_loss, self.seed, additional_run_info_
 
         if isinstance(loss, dict):
             loss_ = loss
@@ -310,15 +392,17 @@ class AbstractEvaluator(object):
             rval_dict['final_queue_element'] = True
 
         self.queue.put(rval_dict)
+        return self.duration, loss_, self.seed, additional_run_info_
 
     def calculate_auxiliary_losses(
         self,
-        Y_valid_pred,
-        Y_test_pred
-    ):
+        Y_valid_pred: np.ndarray,
+        Y_test_pred: np.ndarray,
+    ) -> Tuple[Optional[float], Optional[float]]:
         if Y_valid_pred is not None:
             if self.y_valid is not None:
-                validation_loss = self._loss(self.y_valid, Y_valid_pred)
+                validation_loss: Optional[Union[float, Dict[str, float]]] = self._loss(
+                    self.y_valid, Y_valid_pred)
                 if isinstance(validation_loss, dict):
                     validation_loss = validation_loss[self.metric.name]
             else:
@@ -328,7 +412,8 @@ class AbstractEvaluator(object):
 
         if Y_test_pred is not None:
             if self.y_test is not None:
-                test_loss = self._loss(self.y_test, Y_test_pred)
+                test_loss: Optional[Union[float, Dict[str, float]]] = self._loss(
+                    self.y_test, Y_test_pred)
                 if isinstance(test_loss, dict):
                     test_loss = test_loss[self.metric.name]
             else:
@@ -339,11 +424,11 @@ class AbstractEvaluator(object):
         return validation_loss, test_loss
 
     def file_output(
-            self,
-            Y_optimization_pred,
-            Y_valid_pred,
-            Y_test_pred
-    ):
+        self,
+        Y_optimization_pred: np.ndarray,
+        Y_valid_pred: np.ndarray,
+        Y_test_pred: np.ndarray,
+    ) -> Tuple[Optional[float], Dict[str, Union[str, int, float, List, Dict, Tuple]]]:
         # Abort if self.Y_optimization is None
         # self.Y_optimization can be None if we use partial-cv, then,
         # obviously no output should be saved.
@@ -351,14 +436,14 @@ class AbstractEvaluator(object):
             return None, {}
 
         # Abort in case of shape misalignment
-        if self.Y_optimization.shape[0] != Y_optimization_pred.shape[0]:
+        if np.shape(self.Y_optimization)[0] != Y_optimization_pred.shape[0]:
             return (
                 1.0,
                 {
                     'error':
                         "Targets %s and prediction %s don't have "
                         "the same length. Probably training didn't "
-                        "finish" % (self.Y_optimization.shape, Y_optimization_pred.shape)
+                        "finish" % (np.shape(self.Y_optimization), Y_optimization_pred.shape)
                  },
             )
 
@@ -389,130 +474,82 @@ class AbstractEvaluator(object):
         if self.disable_file_output is False:
             self.disable_file_output = []
 
+        # Here onwards, the self.disable_file_output can be treated as a list
+        self.disable_file_output = cast(List, self.disable_file_output)
+
         # This file can be written independently of the others down bellow
         if ('y_optimization' not in self.disable_file_output):
             if self.output_y_hat_optimization:
-                try:
-                    os.makedirs(self.backend.output_directory)
-                except OSError:
-                    pass
                 self.backend.save_targets_ensemble(self.Y_optimization)
 
-        # The other four files have to be written together, meaning we start
-        # writing them just after acquiring the locks for all of them.
-        # But first we have to check which files have to be written.
-        write_tasks = []
+        models: Optional[BaseEstimator] = None
+        if hasattr(self, 'models'):
+            if len(self.models) > 0 and self.models[0] is not None:  # type: ignore[attr-defined]
+                if ('models' not in self.disable_file_output):
 
-        # File 1 of 5: model
-        if ('model' not in self.disable_file_output):
-            if os.path.exists(self.backend.get_model_dir()):
-                file_path = self.backend.get_model_path(
-                    self.seed, self.num_run, self.budget)
-                write_tasks.append(
-                    WriteTask(
-                        lock=lockfile.LockFile(file_path),
-                        writer=self.backend.save_model,
-                        args=(self.model, file_path)
-                        ))
+                    if self.task_type in CLASSIFICATION_TASKS:
+                        models = VotingClassifier(estimators=None, voting='soft', )
+                    else:
+                        models = VotingRegressor(estimators=None)
+                    # Mypy cannot understand hasattr yet
+                    models.estimators_ = self.models  # type: ignore[attr-defined]
 
-        # File 2 of 5: predictions
-        if ('y_optimization' not in self.disable_file_output):
-            file_path = self.backend.get_prediction_output_path(
-                'ensemble', self.seed, self.num_run, self.budget)
-            write_tasks.append(
-                WriteTask(
-                    lock=lockfile.LockFile(file_path),
-                    writer=self.backend.save_predictions_as_npy,
-                    args=(Y_optimization_pred, file_path)
-                    ))
-
-        # File 3 of 5: validation predictions
-        if Y_valid_pred is not None:
-            file_path = self.backend.get_prediction_output_path(
-                'valid', self.seed, self.num_run, self.budget)
-            write_tasks.append(
-                WriteTask(
-                    lock=lockfile.LockFile(file_path),
-                    writer=self.backend.save_predictions_as_npy,
-                    args=(Y_valid_pred, file_path)
-                    ))
-
-        # File 4 of 5: test predictions
-        if Y_test_pred is not None:
-            file_path = self.backend.get_prediction_output_path(
-                'test', self.seed, self.num_run, self.budget)
-            write_tasks.append(
-                WriteTask(
-                    lock=lockfile.LockFile(file_path),
-                    writer=self.backend.save_predictions_as_npy,
-                    args=(Y_test_pred, file_path)
-                    ))
-
-        # File 5 of 5: ensemble of models in case of cross-validation
-        if hasattr(self, 'models') and len(self.models) > 0 and self.models[0] is not None:
-            if ('models' not in self.disable_file_output):
-
-                if self.task_type in CLASSIFICATION_TASKS:
-                    models = VotingClassifier(estimators=None, voting='soft', )
-                else:
-                    models = VotingRegressor(estimators=None)
-                models.estimators_ = self.models
-
-                if os.path.exists(self.backend.get_cv_model_dir()):
-                    file_path = self.backend.get_cv_model_path(
-                        self.seed, self.num_run, self.budget)
-                    write_tasks.append(
-                        WriteTask(
-                            lock=lockfile.LockFile(file_path),
-                            writer=self.backend.save_model,
-                            args=(models, file_path)
-                            ))
-
-        # We then acquire the locks one by one in a stubborn fashion, i.e. if a file is
-        # already locked, we keep probing it until it is unlocked. This will NOT create a
-        # race condition with _delete_non_candidate_models() since this function doesn't
-        # acquire the locks in this stubborn way. The delete function releases all the
-        # locks and aborts the acquision process as soon as it finds a locked file.
-        for wt in write_tasks:
-            while True:
-                try:
-                    wt.lock.acquire()
-                    break
-                except lockfile.AlreadyLocked:
-                    time.sleep(.1)
-                    continue
-                except Exception as e:
-                    raise RuntimeError('Failed to lock %s due to %s' % (wt.lock, e))
-
-        # At this point we are good to write the files
-        for wt in write_tasks:
-            wt.writer(*wt.args)
-
-        # And finally release the locks
-        for wt in write_tasks:
-            wt.lock.release()
+        self.backend.save_numrun_to_dir(
+            seed=self.seed,
+            idx=self.num_run,
+            budget=self.budget,
+            model=self.model if 'model' not in self.disable_file_output else None,
+            cv_model=models if 'cv_model' not in self.disable_file_output else None,
+            ensemble_predictions=(
+                Y_optimization_pred if 'y_optimization' not in self.disable_file_output else None
+            ),
+            valid_predictions=(
+                Y_valid_pred if 'y_valid' not in self.disable_file_output else None
+            ),
+            test_predictions=(
+                Y_test_pred if 'y_test' not in self.disable_file_output else None
+            ),
+        )
 
         return None, {}
 
-    def _predict_proba(self, X, model, task_type, Y_train):
-        def send_warnings_to_log(message, category, filename, lineno,
-                                 file=None, line=None):
+    def _predict_proba(self, X: np.ndarray, model: BaseEstimator,
+                       task_type: int, Y_train: Optional[np.ndarray] = None,
+                       ) -> np.ndarray:
+        def send_warnings_to_log(
+            message: Union[Warning, str],
+            category: Type[Warning],
+            filename: str,
+            lineno: int,
+            file: Optional[TextIO] = None,
+            line: Optional[str] = None,
+        ) -> None:
             self.logger.debug('%s:%s: %s:%s' %
-                              (filename, lineno, category.__name__, message))
+                              (filename, lineno, str(category), message))
             return
 
         with warnings.catch_warnings():
             warnings.showwarning = send_warnings_to_log
             Y_pred = model.predict_proba(X, batch_size=1000)
 
+        if Y_train is None:
+            raise ValueError("Y_train is required for classification problems")
+
         Y_pred = self._ensure_prediction_array_sizes(Y_pred, Y_train)
         return Y_pred
 
-    def _predict_regression(self, X, model, task_type, Y_train=None):
-        def send_warnings_to_log(message, category, filename, lineno,
-                                 file=None, line=None):
+    def _predict_regression(self, X: np.ndarray, model: BaseEstimator,
+                            task_type: int, Y_train: Optional[np.ndarray] = None) -> np.ndarray:
+        def send_warnings_to_log(
+            message: Union[Warning, str],
+            category: Type[Warning],
+            filename: str,
+            lineno: int,
+            file: Optional[TextIO] = None,
+            line: Optional[str] = None,
+        ) -> None:
             self.logger.debug('%s:%s: %s:%s' %
-                              (filename, lineno, category.__name__, message))
+                              (filename, lineno, str(category), message))
             return
 
         with warnings.catch_warnings():
@@ -524,7 +561,8 @@ class AbstractEvaluator(object):
 
         return Y_pred
 
-    def _ensure_prediction_array_sizes(self, prediction, Y_train):
+    def _ensure_prediction_array_sizes(self, prediction: np.ndarray, Y_train: np.ndarray
+                                       ) -> np.ndarray:
         num_classes = self.datamanager.info['label_num']
 
         if self.task_type == MULTICLASS_CLASSIFICATION and \
