@@ -1,13 +1,30 @@
 from abc import ABCMeta
+from typing import Dict, Union
+
+from ConfigSpace import Configuration
 
 import numpy as np
-from ConfigSpace import Configuration
+
+import scipy.sparse
+
 from sklearn.pipeline import Pipeline
 from sklearn.utils.validation import check_random_state
 
 from .components.base import AutoSklearnChoice, AutoSklearnComponent
 import autosklearn.pipeline.create_searchspace_util
 
+DATASET_PROPERTIES_TYPE = Dict[str, Union[str, int, bool]]
+PIPELINE_DATA_DTYPE = Union[
+    np.ndarray,
+    scipy.sparse.bsr_matrix,
+    scipy.sparse.coo_matrix,
+    scipy.sparse.csc_matrix,
+    scipy.sparse.csr_matrix,
+    scipy.sparse.dia_matrix,
+    scipy.sparse.dok_matrix,
+    scipy.sparse.lil_matrix,
+]
+
 
 class BasePipeline(Pipeline):
     """Base class for all pipeline objects.
@@ -28,7 +45,7 @@ class BasePipeline(Pipeline):
             dataset_properties is not None else {}
 
         if steps is None:
-            self.steps = self._get_pipeline_steps()
+            self.steps = self._get_pipeline_steps(dataset_properties=dataset_properties)
         else:
             self.steps = steps
 
@@ -98,10 +115,9 @@ class BasePipeline(Pipeline):
             fit_params = {}
         fit_params = {key.replace(":", "__"): value for key, value in
                       fit_params.items()}
-        Xt, fit_params = self._fit(X, y, **fit_params)
-        if fit_params is None:
-            fit_params = {}
-        return Xt, fit_params
+        fit_params_steps = self._check_fit_params(**fit_params)
+        Xt = self._fit(X, y, **fit_params_steps)
+        return Xt, fit_params_steps[self.steps[-1][0]]
 
     def fit_estimator(self, X, y, **fit_params):
         fit_params = {key.replace(":", "__"): value for key, value in
@@ -173,7 +189,7 @@ class BasePipeline(Pipeline):
                 return y
 
     def set_hyperparameters(self, configuration, init_params=None):
-        self.configuration = configuration
+        self.config = configuration
 
         for node_idx, n_ in enumerate(self.steps):
             node_name, node = n_
@@ -207,6 +223,10 @@ class BasePipeline(Pipeline):
             else:
                 raise NotImplementedError('Not supported yet!')
 
+        # In-code check to make sure init params
+        # is checked after pipeline creation
+        self._check_init_params_honored(init_params)
+
         return self
 
     def get_hyperparameter_search_space(self, dataset_properties=None):
@@ -345,6 +365,58 @@ class BasePipeline(Pipeline):
 
         return cs
 
+    def _check_init_params_honored(self, init_params):
+        """
+        Makes sure that init params is honored at the implementation level
+        """
+        if init_params is None or len(init_params) < 1:
+            # None/empty dict, so no further check required
+            return
+
+        # There is the scenario, where instance is passed as an argument to the init_params
+        # 'instance': '{"task_id": "73543c4a360aa24498c0967fbc2f926b"}'}
+        # coming from smac instance. Remove this key to make the testing stricter
+        init_params.pop('instance', None)
+
+        for key, value in init_params.items():
+
+            if ':' not in key:
+                raise ValueError("Unsupported argument to init_params {}."
+                                 "When using init_params, a hierarchical format like "
+                                 "node_name:parameter must be provided.".format(key)
+                                 )
+            node_name = key.split(':', 1)[0]
+            if node_name not in self.named_steps.keys():
+                raise ValueError("The current node name specified via key={} of init_params "
+                                 "is not valid. Valid node names are {}".format(
+                                     key,
+                                     self.named_steps.keys()
+                                 )
+                                 )
+                continue
+            variable_name = key.split(':')[1]
+            node = self.named_steps[node_name]
+            if isinstance(node, BasePipeline):
+                # If dealing with a sub pipe,
+                # Call the child _check_init_params_honored with the updated config
+                node._check_init_params_honored(
+                    {
+                        key.replace('%s:' % node_name, '', 1): value
+                    }
+                )
+                continue
+
+            if isinstance(node, AutoSklearnComponent):
+                node_dict = vars(node)
+            elif isinstance(node, AutoSklearnChoice):
+                node_dict = vars(node.choice)
+            else:
+                raise ValueError("Unsupported node type {}".format(type(node)))
+
+            if variable_name not in node_dict or node_dict[variable_name] != value:
+                raise ValueError("Cannot properly set the pair {}->{} via init_params"
+                                 "".format(key, value))
+
     def __repr__(self):
         class_name = self.__class__.__name__
 
@@ -385,7 +457,7 @@ class BasePipeline(Pipeline):
 
         return rval
 
-    def _get_pipeline_steps(self):
+    def _get_pipeline_steps(self, dataset_properties):
         raise NotImplementedError()
 
     def _get_estimator_hyperparameter_name(self):
