@@ -1,26 +1,37 @@
+import copy
 import json
+import logging
+import multiprocessing
 import os
 import time
 import traceback
+import typing
 import warnings
 
-import numpy as np
+import dask.distributed
 import pynisher
 
 from smac.facade.smac_ac_facade import SMAC4AC
+from smac.intensification.simple_intensifier import SimpleIntensifier
+from smac.intensification.intensification import Intensifier
 from smac.runhistory.runhistory2epm import RunHistory2EPM4LogCost
 from smac.scenario.scenario import Scenario
-from smac.optimizer import pSMAC
+from smac.tae.serial_runner import SerialRunner
+from smac.tae.dask_runner import DaskParallelRunner
+from smac.callbacks import IncorporateRunResultCallback
 
 
 import autosklearn.metalearning
 from autosklearn.constants import MULTILABEL_CLASSIFICATION, \
     BINARY_CLASSIFICATION, TASK_TYPES_TO_STRING, CLASSIFICATION_TASKS, \
-    REGRESSION_TASKS, MULTICLASS_CLASSIFICATION, REGRESSION
+    MULTICLASS_CLASSIFICATION, REGRESSION, MULTIOUTPUT_REGRESSION
+from autosklearn.ensemble_builder import EnsembleBuilderManager
 from autosklearn.metalearning.mismbo import suggest_via_metalearning
 from autosklearn.data.abstract_data_manager import AbstractDataManager
 from autosklearn.evaluation import ExecuteTaFuncWithQueue, get_cost_of_crash
-from autosklearn.util.logging_ import get_logger
+from autosklearn.util.logging_ import get_named_client_logger
+from autosklearn.util.parallel import preload_modules
+from autosklearn.util.pipeline import parse_include_exclude_components
 from autosklearn.metalearning.metalearning.meta_base import MetaBase
 from autosklearn.metalearning.metafeatures.metafeatures import \
     calculate_all_metafeatures_with_labels, calculate_all_metafeatures_encoded_labels
@@ -31,10 +42,11 @@ EXCLUDE_META_FEATURES_CLASSIFICATION = {
     'LandmarkDecisionTree',
     'LandmarkLDA',
     'LandmarkNaiveBayes',
+    'LandmarkRandomNodeLearner',
     'PCAFractionOfComponentsFor95PercentVariance',
     'PCAKurtosisFirstPC',
     'PCASkewnessFirstPC',
-    'PCA'
+    'PCA',
 }
 
 EXCLUDE_META_FEATURES_REGRESSION = {
@@ -58,57 +70,74 @@ EXCLUDE_META_FEATURES_REGRESSION = {
 }
 
 
+def get_send_warnings_to_logger(logger):
+    def _send_warnings_to_log(message, category, filename, lineno, file, line):
+        logger.debug('%s:%s: %s:%s', filename, lineno, category.__name__, message)
+    return _send_warnings_to_log
+
+
 # metalearning helpers
 def _calculate_metafeatures(data_feat_type, data_info_task, basename,
-                            x_train, y_train, watcher, logger):
-    # == Calculate metafeatures
-    task_name = 'CalculateMetafeatures'
-    watcher.start_task(task_name)
-    categorical = [True if feat_type.lower() in ['categorical'] else False
-                   for feat_type in data_feat_type]
+                            x_train, y_train, watcher, logger_):
+    with warnings.catch_warnings():
+        warnings.showwarning = get_send_warnings_to_logger(logger_)
 
-    EXCLUDE_META_FEATURES = EXCLUDE_META_FEATURES_CLASSIFICATION \
-        if data_info_task in CLASSIFICATION_TASKS else EXCLUDE_META_FEATURES_REGRESSION
+        # == Calculate metafeatures
+        task_name = 'CalculateMetafeatures'
+        watcher.start_task(task_name)
 
-    if data_info_task in [MULTICLASS_CLASSIFICATION, BINARY_CLASSIFICATION,
-                          MULTILABEL_CLASSIFICATION, REGRESSION]:
-        logger.info('Start calculating metafeatures for %s', basename)
-        result = calculate_all_metafeatures_with_labels(
+        categorical = {col: True if feat_type.lower() == 'categorical' else False
+                       for col, feat_type in data_feat_type.items()}
+
+        EXCLUDE_META_FEATURES = EXCLUDE_META_FEATURES_CLASSIFICATION \
+            if data_info_task in CLASSIFICATION_TASKS else EXCLUDE_META_FEATURES_REGRESSION
+
+        if data_info_task in [MULTICLASS_CLASSIFICATION, BINARY_CLASSIFICATION,
+                              MULTILABEL_CLASSIFICATION, REGRESSION,
+                              MULTIOUTPUT_REGRESSION]:
+            logger_.info('Start calculating metafeatures for %s', basename)
+            result = calculate_all_metafeatures_with_labels(
+                x_train, y_train, categorical=categorical,
+                dataset_name=basename,
+                dont_calculate=EXCLUDE_META_FEATURES, logger=logger_)
+            for key in list(result.metafeature_values.keys()):
+                if result.metafeature_values[key].type_ != 'METAFEATURE':
+                    del result.metafeature_values[key]
+
+        else:
+            result = None
+            logger_.info('Metafeatures not calculated')
+        watcher.stop_task(task_name)
+        logger_.info(
+            'Calculating Metafeatures (categorical attributes) took %5.2f',
+            watcher.wall_elapsed(task_name))
+        return result
+
+
+def _calculate_metafeatures_encoded(data_feat_type, basename, x_train, y_train, watcher,
+                                    task, logger_):
+    with warnings.catch_warnings():
+        warnings.showwarning = get_send_warnings_to_logger(logger_)
+
+        EXCLUDE_META_FEATURES = EXCLUDE_META_FEATURES_CLASSIFICATION \
+            if task in CLASSIFICATION_TASKS else EXCLUDE_META_FEATURES_REGRESSION
+
+        task_name = 'CalculateMetafeaturesEncoded'
+        watcher.start_task(task_name)
+        categorical = {col: True if feat_type.lower() == 'categorical' else False
+                       for col, feat_type in data_feat_type.items()}
+
+        result = calculate_all_metafeatures_encoded_labels(
             x_train, y_train, categorical=categorical,
-            dataset_name=basename,
-            dont_calculate=EXCLUDE_META_FEATURES, )
+            dataset_name=basename, dont_calculate=EXCLUDE_META_FEATURES, logger=logger_)
         for key in list(result.metafeature_values.keys()):
             if result.metafeature_values[key].type_ != 'METAFEATURE':
                 del result.metafeature_values[key]
-
-    else:
-        result = None
-        logger.info('Metafeatures not calculated')
-    watcher.stop_task(task_name)
-    logger.info(
-        'Calculating Metafeatures (categorical attributes) took %5.2f',
-        watcher.wall_elapsed(task_name))
-    return result
-
-
-def _calculate_metafeatures_encoded(basename, x_train, y_train, watcher,
-                                    task, logger):
-    EXCLUDE_META_FEATURES = EXCLUDE_META_FEATURES_CLASSIFICATION \
-        if task in CLASSIFICATION_TASKS else EXCLUDE_META_FEATURES_REGRESSION
-
-    task_name = 'CalculateMetafeaturesEncoded'
-    watcher.start_task(task_name)
-    result = calculate_all_metafeatures_encoded_labels(
-        x_train, y_train, categorical=[False] * x_train.shape[1],
-        dataset_name=basename, dont_calculate=EXCLUDE_META_FEATURES)
-    for key in list(result.metafeature_values.keys()):
-        if result.metafeature_values[key].type_ != 'METAFEATURE':
-            del result.metafeature_values[key]
-    watcher.stop_task(task_name)
-    logger.info(
-        'Calculating Metafeatures (encoded attributes) took %5.2fsec',
-        watcher.wall_elapsed(task_name))
-    return result
+        watcher.stop_task(task_name)
+        logger_.info(
+            'Calculating Metafeatures (encoded attributes) took %5.2fsec',
+            watcher.wall_elapsed(task_name))
+        return result
 
 
 def _get_metalearning_configurations(meta_base, basename, metric,
@@ -124,7 +153,8 @@ def _get_metalearning_configurations(meta_base, basename, metric,
             meta_base, basename, metric,
             task,
             is_sparse == 1,
-            initial_configurations_via_metalearning
+            initial_configurations_via_metalearning,
+            logger=logger,
         )
     except Exception as e:
         logger.error("Error getting metalearning configurations!")
@@ -152,12 +182,15 @@ def get_smac_object(
     seed,
     ta,
     ta_kwargs,
-    backend,
     metalearning_configurations,
+    n_jobs,
+    dask_client,
 ):
-    scenario_dict['input_psmac_dirs'] = backend.get_smac_output_glob(
-        smac_run_id=seed if not scenario_dict['shared-model'] else '*',
-    )
+    if len(scenario_dict['instances']) > 1:
+        intensifier = Intensifier
+    else:
+        intensifier = SimpleIntensifier
+
     scenario = Scenario(scenario_dict)
     if len(metalearning_configurations) > 0:
         default_config = scenario.cs.get_default_configuration()
@@ -173,6 +206,9 @@ def get_smac_object(
         tae_runner_kwargs=ta_kwargs,
         initial_configurations=initial_configurations,
         run_id=seed,
+        intensifier=intensifier,
+        dask_client=dask_client,
+        n_jobs=n_jobs,
     )
 
 
@@ -184,7 +220,11 @@ class AutoMLSMBO(object):
                  func_eval_time_limit,
                  memory_limit,
                  metric,
-                 watcher, start_num_run=1,
+                 watcher,
+                 n_jobs,
+                 dask_client: dask.distributed.Client,
+                 port: int,
+                 start_num_run=1,
                  data_memory_limit=None,
                  num_metalearning_cfgs=25,
                  config_file=None,
@@ -192,14 +232,18 @@ class AutoMLSMBO(object):
                  metadata_directory=None,
                  resampling_strategy='holdout',
                  resampling_strategy_args=None,
-                 shared_mode=False,
                  include_estimators=None,
                  exclude_estimators=None,
                  include_preprocessors=None,
                  exclude_preprocessors=None,
                  disable_file_output=False,
                  smac_scenario_args=None,
-                 get_smac_object_callback=None):
+                 get_smac_object_callback=None,
+                 scoring_functions=None,
+                 pynisher_context='spawn',
+                 ensemble_callback: typing.Optional[EnsembleBuilderManager] = None,
+                 trials_callback: typing.Optional[IncorporateRunResultCallback] = None
+                 ):
         super(AutoMLSMBO, self).__init__()
         # data related
         self.dataset_name = dataset_name
@@ -207,10 +251,15 @@ class AutoMLSMBO(object):
         self.metric = metric
         self.task = None
         self.backend = backend
+        self.port = port
 
         # the configuration space
         self.config_space = config_space
 
+        # the number of parallel workers/jobs
+        self.n_jobs = n_jobs
+        self.dask_client = dask_client
+
         # Evaluation
         self.resampling_strategy = resampling_strategy
         if resampling_strategy_args is None:
@@ -229,7 +278,6 @@ class AutoMLSMBO(object):
         self.seed = seed
         self.metadata_directory = metadata_directory
         self.start_num_run = start_num_run
-        self.shared_mode = shared_mode
         self.include_estimators = include_estimators
         self.exclude_estimators = exclude_estimators
         self.include_preprocessors = include_preprocessors
@@ -237,15 +285,22 @@ class AutoMLSMBO(object):
         self.disable_file_output = disable_file_output
         self.smac_scenario_args = smac_scenario_args
         self.get_smac_object_callback = get_smac_object_callback
+        self.scoring_functions = scoring_functions
+
+        self.pynisher_context = pynisher_context
+
+        self.ensemble_callback = ensemble_callback
+        self.trials_callback = trials_callback
 
         dataset_name_ = "" if dataset_name is None else dataset_name
         logger_name = '%s(%d):%s' % (self.__class__.__name__, self.seed, ":" + dataset_name_)
-        self.logger = get_logger(logger_name)
-
-    def _send_warnings_to_log(self, message, category, filename, lineno,
-                              file=None, line=None):
-        self.logger.debug('%s:%s: %s:%s', filename, lineno, category.__name__,
-                          message)
+        if port is None:
+            self.logger = logging.getLogger(__name__)
+        else:
+            self.logger = get_named_client_logger(
+                name=logger_name,
+                port=self.port,
+            )
 
     def reset_data_manager(self, max_mem=None):
         if max_mem is None:
@@ -279,58 +334,53 @@ class AutoMLSMBO(object):
 
         return metalearning_configurations
 
-    def _calculate_metafeatures(self):
-        with warnings.catch_warnings():
-            warnings.showwarning = self._send_warnings_to_log
-
-            meta_features = _calculate_metafeatures(
+    def _calculate_metafeatures_with_limits(self, time_limit):
+        res = None
+        time_limit = max(time_limit, 1)
+        try:
+            context = multiprocessing.get_context(self.pynisher_context)
+            preload_modules(context)
+            safe_mf = pynisher.enforce_limits(mem_in_mb=self.memory_limit,
+                                              wall_time_in_s=int(time_limit),
+                                              grace_period_in_s=30,
+                                              context=context,
+                                              logger=self.logger)(
+                _calculate_metafeatures)
+            res = safe_mf(
                 data_feat_type=self.datamanager.feat_type,
                 data_info_task=self.datamanager.info['task'],
                 x_train=self.datamanager.data['X_train'],
                 y_train=self.datamanager.data['Y_train'],
                 basename=self.dataset_name,
                 watcher=self.watcher,
-                logger=self.logger)
-            return meta_features
-
-    def _calculate_metafeatures_with_limits(self, time_limit):
-        res = None
-        time_limit = max(time_limit, 1)
-        try:
-            safe_mf = pynisher.enforce_limits(mem_in_mb=self.memory_limit,
-                                              wall_time_in_s=int(time_limit),
-                                              grace_period_in_s=30,
-                                              logger=self.logger)(
-                self._calculate_metafeatures)
-            res = safe_mf()
+                logger_=self.logger
+            )
         except Exception as e:
             self.logger.error('Error getting metafeatures: %s', str(e))
 
         return res
 
-    def _calculate_metafeatures_encoded(self):
-        with warnings.catch_warnings():
-            warnings.showwarning = self._send_warnings_to_log
-
-            meta_features_encoded = _calculate_metafeatures_encoded(
-                self.dataset_name,
-                self.datamanager.data['X_train'],
-                self.datamanager.data['Y_train'],
-                self.watcher,
-                self.datamanager.info['task'],
-                self.logger)
-            return meta_features_encoded
-
     def _calculate_metafeatures_encoded_with_limits(self, time_limit):
         res = None
         time_limit = max(time_limit, 1)
         try:
+            context = multiprocessing.get_context(self.pynisher_context)
+            preload_modules(context)
             safe_mf = pynisher.enforce_limits(mem_in_mb=self.memory_limit,
                                               wall_time_in_s=int(time_limit),
                                               grace_period_in_s=30,
+                                              context=context,
                                               logger=self.logger)(
-                self._calculate_metafeatures_encoded)
-            res = safe_mf()
+                _calculate_metafeatures_encoded)
+            res = safe_mf(
+                data_feat_type=self.datamanager.feat_type,
+                task=self.datamanager.info['task'],
+                x_train=self.datamanager.data['X_train'],
+                y_train=self.datamanager.data['Y_train'],
+                basename=self.dataset_name,
+                watcher=self.watcher,
+                logger_=self.logger
+            )
         except Exception as e:
             self.logger.error('Error getting metafeatures (encoded) : %s',
                               str(e))
@@ -368,48 +418,30 @@ class AutoMLSMBO(object):
         # evaluator, which takes into account that a run can be killed prior
         # to the model being fully fitted; thus putting intermediate results
         # into a queue and querying them once the time is over
-        exclude = dict()
-        include = dict()
-        if self.include_preprocessors is not None and self.exclude_preprocessors is not None:
-            raise ValueError('Cannot specify include_preprocessors and '
-                             'exclude_preprocessors.')
-        elif self.include_preprocessors is not None:
-            include['feature_preprocessor'] = self.include_preprocessors
-        elif self.exclude_preprocessors is not None:
-            exclude['feature_preprocessor'] = self.exclude_preprocessors
+        include, exclude = parse_include_exclude_components(
+            task=self.task,
+            include_estimators=self.include_estimators,
+            exclude_estimators=self.exclude_estimators,
+            include_preprocessors=self.include_preprocessors,
+            exclude_preprocessors=self.exclude_preprocessors,
+        )
 
-        if self.include_estimators is not None and self.exclude_estimators is not None:
-            raise ValueError('Cannot specify include_estimators and '
-                             'exclude_estimators.')
-        elif self.include_estimators is not None:
-            if self.task in CLASSIFICATION_TASKS:
-                include['classifier'] = self.include_estimators
-            elif self.task in REGRESSION_TASKS:
-                include['regressor'] = self.include_estimators
-            else:
-                raise ValueError(self.task)
-        elif self.exclude_estimators is not None:
-            if self.task in CLASSIFICATION_TASKS:
-                exclude['classifier'] = self.exclude_estimators
-            elif self.task in REGRESSION_TASKS:
-                exclude['regressor'] = self.exclude_estimators
-            else:
-                raise ValueError(self.task)
-
-        ta = ExecuteTaFuncWithQueue
         ta_kwargs = dict(
-            backend=self.backend,
+            backend=copy.deepcopy(self.backend),
             autosklearn_seed=seed,
             resampling_strategy=self.resampling_strategy,
             initial_num_run=num_run,
-            logger=self.logger,
             include=include,
             exclude=exclude,
             metric=self.metric,
             memory_limit=self.memory_limit,
             disable_file_output=self.disable_file_output,
+            scoring_functions=self.scoring_functions,
+            port=self.port,
+            pynisher_context=self.pynisher_context,
             **self.resampling_strategy_args
         )
+        ta = ExecuteTaFuncWithQueue
 
         startup_time = self.watcher.wall_elapsed(self.dataset_name)
         total_walltime_limit = self.total_walltime_limit - startup_time - 5
@@ -422,7 +454,6 @@ class AutoMLSMBO(object):
             'memory_limit': self.memory_limit,
             'output-dir': self.backend.get_smac_output_directory(),
             'run_obj': 'quality',
-            'shared-model': self.shared_mode,
             'wallclock_limit': total_walltime_limit,
             'cost_for_crash': self.worst_possible_result,
         }
@@ -460,29 +491,30 @@ class AutoMLSMBO(object):
             'seed': seed,
             'ta': ta,
             'ta_kwargs': ta_kwargs,
-            'backend': self.backend,
             'metalearning_configurations': metalearning_configurations,
+            'n_jobs': self.n_jobs,
+            'dask_client': self.dask_client,
         }
         if self.get_smac_object_callback is not None:
             smac = self.get_smac_object_callback(**smac_args)
         else:
             smac = get_smac_object(**smac_args)
 
-        smac.optimize()
+        if self.ensemble_callback is not None:
+            smac.register_callback(self.ensemble_callback)
+        if self.trials_callback is not None:
+            smac.register_callback(self.trials_callback)
 
-        # Patch SMAC to read in data from parallel runs after the last
-        # function evaluation
-        if self.shared_mode:
-            pSMAC.read(
-                run_history=smac.solver.runhistory,
-                output_dirs=smac.solver.scenario.input_psmac_dirs,
-                configuration_space=smac.solver.config_space,
-                logger=smac.solver.logger,
-            )
+        smac.optimize()
 
         self.runhistory = smac.solver.runhistory
         self.trajectory = smac.solver.intensifier.traj_logger.trajectory
-        self._budget_type = smac.solver.intensifier.tae_runner.budget_type
+        if isinstance(smac.solver.tae_runner, DaskParallelRunner):
+            self._budget_type = smac.solver.tae_runner.single_worker.budget_type
+        elif isinstance(smac.solver.tae_runner, SerialRunner):
+            self._budget_type = smac.solver.tae_runner.budget_type
+        else:
+            raise NotImplementedError(type(smac.solver.tae_runner))
 
         return self.runhistory, self.trajectory, self._budget_type
 
@@ -542,7 +574,7 @@ class AutoMLSMBO(object):
 
                 self.logger.info('Metadata directory: %s',
                                  self.metadata_directory)
-                meta_base = MetaBase(self.config_space, self.metadata_directory)
+                meta_base = MetaBase(self.config_space, self.metadata_directory, self.logger)
 
                 metafeature_calculation_time_limit = int(
                     self.total_walltime_limit / 4)
@@ -564,8 +596,7 @@ class AutoMLSMBO(object):
                     meta_features_encoded = None
                 else:
                     with warnings.catch_warnings():
-                        warnings.showwarning = self._send_warnings_to_log
-                        self.datamanager.perform1HotEncoding()
+                        warnings.showwarning = get_send_warnings_to_logger(self.logger)
                     meta_features_encoded = \
                         self._calculate_metafeatures_encoded_with_limits(
                             metafeature_calculation_time_limit)
@@ -589,7 +620,7 @@ class AutoMLSMBO(object):
                                             inplace=True)
 
                     with warnings.catch_warnings():
-                        warnings.showwarning = self._send_warnings_to_log
+                        warnings.showwarning = get_send_warnings_to_logger(self.logger)
                         metalearning_configurations = self.collect_metalearning_suggestions(
                             meta_base)
                     if metalearning_configurations is None:
@@ -607,8 +638,6 @@ class AutoMLSMBO(object):
                     for meta_feature_name in all_metafeatures.columns:
                         meta_features_list.append(
                             meta_features[meta_feature_name].value)
-                    meta_features_list = np.array(meta_features_list).reshape(
-                        (1, -1))
                     self.logger.info(list(meta_features_dict.keys()))
 
             else:
@@ -619,6 +648,5 @@ class AutoMLSMBO(object):
         else:
             meta_features = None
         if meta_features is None:
-            meta_features_list = []
             metalearning_configurations = []
         return metalearning_configurations
